import requests
import soundcard as sc
import soundfile as sf
import io


def speaker(filename: str, seconds):
    """
    Records the device's speaker.

    :param filename: Name and directory of the audio file written.
    :param seconds: Duration to record (seconds).
    """
    with sc.get_microphone(
        id=str(sc.default_speaker().name), include_loopback=True
    ).recorder(samplerate=44100) as speaker:
        print(f"Recording speaker for {seconds} seconds...")

        data = speaker.record(numframes=44100 * seconds)
        sf.write(file=filename, data=data, samplerate=44100)
    return filename


def microphone(filename: str, seconds):
    """
    Records the device's device.

    :param filename: Name and directory of the audio file written.
    :param seconds: Duration to record (seconds).
    """
    with sc.get_microphone(
        id=str(sc.default_microphone().name), include_loopback=True
    ).recorder(samplerate=44100) as mic:
        print(f"Recording microphone for {seconds} seconds...")
        data = mic.record(numframes=44100 * seconds)
        sf.write(file=filename, data=data, samplerate=44100)
    return filename


def url(url: str, filename: str):
    """
    Downloads audio from the provided URL.

    :param url: URL of the audio file.
    :param filename: Name and directory of the audio file written.
    :param seconds: Duration to record (seconds).
    """
    print(f"Downloading audio from URL...")
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.246'}

    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()  # Raises an HTTPError if code is 4XX/5XX

    except requests.exceptions.HTTPError as http_err:
        print(f'HTTP error occurred: {http_err}')
        return None
    except Exception as err:
        print(f'An error occurred: {err}')
        return None

    # Ensure the response content is in a suitable format for soundfile
    data, samplerate = sf.read(io.BytesIO(response.content))
    sf.write(file=filename, data=data, samplerate=samplerate)
    return filename


import asyncio
import csv
from datetime import datetime
import random
from rich.progress import MofNCompleteColumn, Progress, TimeElapsedColumn
from shazamio import Shazam
import soundfile as sf
import tempfile
import time
import os


def write_csv(file_name: str, data_rows: list):
    """
    Writes the found song's data into a CSV file.

    :param file_name: Name for csv file
    :param data_rows: Includes timestamp, track_title, artist and album_cover.
    """
    header = ["Timestamp", "Track", "Artist", "Album Cover"]
    csv_file = f"{file_name}.csv"
    file_exists = os.path.isfile(csv_file)

    with open(csv_file, mode="a", newline="", encoding="utf-8") as file:
        writer = csv.writer(file)
        if not file_exists:
            writer.writerow(header)
        writer.writerow(data_rows)


def extract_track_info(out):
    track_title = out["track"]["title"]
    artist = out["track"]["subtitle"]
    if "images" in out["track"]:
        album_cover_hq = (
            out["track"]
            .get("images", {})
            .get("coverart", "")
            .replace("/400x400cc.jpg", "/1400x1400cc.png")
        )
    else:
        album_cover_hq = None
    return track_title, artist, album_cover_hq


def print_track_info(track_info):
    print(f"Track: {track_info[0]}")
    print(f"Artist: {track_info[1]}")
    print(f"Album Cover: {track_info[2]}")


async def identify_audio(
    audio_file: str, json=False, write=False, timestamp=None, file_name=None
):
    """
    Identifies the song of a given audio file.

    file_name is set to "{audio_file}" if none is provided.

    :param audio_file: Path to audio file.
    :param json: If true, function returns the whole shazamio output as JSON.
    :param write: If true, output is written in CSV file. Timestamp is required if write is True.
    :param timestamp: Time at which the song was recognized. Required for write.
    :param file_name (Optional): Name for csv file Only used if write is True.
    """
    if not os.path.exists(audio_file):
        print(f"Error: File '{audio_file}' not found.")
        return
    # Shazam's maximum API call limit is 20 requests per minute.
    # To avoid API call limits, Shazamio needs to a proxy.
    shazam = Shazam()
    out = await shazam.recognize(data=audio_file, proxy=None)

    if "track" not in out:
        print("No matches found.")
        return

    if json:
        print(out)
        return out

    track_info = extract_track_info(out)
    print_track_info(track_info)

    if write:
        file_name = file_name or os.path.splitext(os.path.basename(audio_file))[0]
        write_csv(file_name, [timestamp, *track_info])


def split_and_identify(audio_file: str, duration: int):
    """
    Splits audio file by duration (seconds) and identifies songs. Writes the csv file automatically.

    :param audio_file: Path to audio file.
    :param duration: Length of each segment (seconds).
    """
    if not os.path.exists(audio_file):
        print(f"Error: File '{audio_file}' not found.")
        return

    file_name = os.path.splitext(os.path.basename(audio_file))[0]
    print(f"Reading {audio_file}...")
    data, samplerate = sf.read(audio_file)
    samples_per_duration = samplerate * duration
    num_segments = len(data) // samples_per_duration

    with Progress(
        TimeElapsedColumn(),
        MofNCompleteColumn(),
        *Progress.get_default_columns(),
    ) as progress:
        task = progress.add_task(
            f"[green]Splitting and identifying audio...", total=num_segments
        )
        # Make each segment an audio file to be used in Shazamio.
        for i in range(num_segments):
            start_idx = i * samples_per_duration
            segment_data = data[start_idx : start_idx + samples_per_duration]
            timestamp = datetime.utcfromtimestamp(start_idx / samplerate).strftime(
                "%H:%M:%S"
            )
            # Temporarily saves audio file, shazams, then deletes.
            with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp_file:
                sf.write(temp_file.name, segment_data, samplerate)
                asyncio.run(
                    identify_audio(
                        temp_file.name,
                        json=False,
                        write=True,
                        timestamp=timestamp,
                        file_name=f"mixtape_{file_name}_{duration}sec",
                    )
                )
                temp_file.close()
                os.remove(temp_file.name)
            progress.update(task, advance=1)
            time.sleep(2)  # Throttle requests to < 20/minute or Error 429.



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyzam",
    version="0.11",
    entry_points={"console_scripts": ["pyzam = pyzam.__main__:main"]},
    author="lukafilipxvic",
    description="A CLI music recognition tool for audio and mixtapes.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lukafilipxvic/Pyzam",
    install_requires=[
        "asyncio",
        "rich",
        "soundfile",
        "shazamio",
        "SoundCard",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3", 
        "License :: OSI Approved :: MIT License",
    ]
)



#! /usr/bin/env python3

"""
Pyzam 0.11
A CLI music recognition tool for audio and mixtapes.
"""

import argparse
import asyncio
from datetime import datetime
from pathlib import Path
from pyzam import identify, record
import shutil
import soundfile as sf
import sys
import tempfile


def _parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog="Pyzam",
        description="CLI music recognition in python.",
    )
    input_group = parser.add_mutually_exclusive_group(required=True)
    run_group = parser.add_mutually_exclusive_group()
    output_group = parser.add_mutually_exclusive_group()

    input_group.add_argument(
        "--input", type=Path, help="detect from the given audio input file"
    )
    input_group.add_argument(
        "-m", "--microphone", action="store_true", help="record audio from microphone"
    )
    input_group.add_argument(
        "-s",
        "--speaker",
        action="store_true",
        help="record audio from speaker (default)",
    )
    input_group.add_argument(
        "-u",
        "--url",
        help="Detect from an audio URL",
    )

    parser.add_argument(
        "-d", "--duration", type=int, default=5, help="audio recording duration (s)"
    )
    run_group.add_argument(
        "--loop", "-l", action="store_true", help="loop music recognition process"
    )
    run_group.add_argument(
        "--mixtape", action="store_true", help="Identifies songs of a mixtape"
    )
    output_group.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="emits the whole shazamio response as raw JSON",
    )
    output_group.add_argument(
        "-w", "--write", action="store_true", help="writes the output as a file"
    )

    return parser


def check_ffmpeg():
    if not shutil.which("ffmpeg"):
        print("[red]Fatal: ffmpeg not found on $PATH[/red]")
        sys.exit(1)


def get_input_file(args, temp_dir) -> Path:
    if args.microphone:
        return record.microphone(
            filename=f"{temp_dir}/pyzam_mic.wav", seconds=args.duration
        )
    elif args.speaker:
        return record.speaker(
            filename=f"{temp_dir}/pyzam_speaker.wav", seconds=args.duration
        )
    elif args.url:
        return record.url(url=args.url,
            filename=f"{temp_dir}/pyzam_url.wav"
        )
    else:
        return args.input


def process_mixtape(input_file, duration):
    song = sf.SoundFile(str(input_file))
    audio_length = song.frames / song.samplerate
    iterations = int(-(-audio_length // duration))
    print(f"Get ready, we are Pyzaming {iterations} times...")
    identify.split_and_identify(str(input_file), duration)


def identify_audio(input_file, args):
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    asyncio.run(
        identify.identify_audio(
            str(input_file), json=args.json, write=args.write, timestamp=current_time
        )
    )


def main() -> None:
    check_ffmpeg()
    parser = _parser()
    args = parser.parse_args()

    # Check if  --input or --url and --loop are used together
    if args.input and args.loop:
        parser.error("--loop is only allowed with --microphone or --speaker")
    if args.url and args.loop:
        parser.error("--loop is only allowed with --microphone or --speaker")

    # Check if --mixtape and (--microphone or --speaker) are used together
    if args.mixtape and (args.microphone or args.speaker):
        parser.error("--mixtape is only allowed with --input or --url")

    temp_dir = tempfile.gettempdir()

    while True:
        input_file = get_input_file(args, temp_dir)
        if args.mixtape:
            process_mixtape(input_file, args.duration)
            break
        identify_audio(input_file, args)
        if not args.loop:
            break


if __name__ == "__main__":
    main()
